import numpy as np
from collections import defaultdict

from .atom_batch import _AtomBatch
from .bond_batch import _BondBatch


class _PlacementsSummary:
    """
    A summary of placement results.

    """

    __slots__ = [
        '_atoms',
        '_atom_infos',
        '_bonds',
        '_bond_infos',
        '_edge_functional_groups',
        '_position_matrices',
        '_num_atoms',
    ]

    def __init__(
        self,
        building_blocks,
        placement_results,
        num_atoms,
        num_previous_placements,
    ):
        """
        Initialize a :class:`._PlacementsSummary` instance.

        Parameters
        ----------
        building_blocks : :class:`iterable` of :class:`.BuildingBlock`
            The building blocks which were placed.

        placement_results : :class:`iterable`
            Holds a :class:`_PlacementResults` instance for each
            building block in `building_blocks`.

        num_atoms : :class:`int`
            The number of atoms in molecule being constructed,
            before this summary is taken into account.

        num_previous_placements : :class:`int`
            The total number of building block placements done
            previously.

        """

        self._atoms = []
        self._atom_infos = []
        self._bonds = []
        self._bond_infos = []
        self._edge_functional_groups = defaultdict(list)
        self._position_matrices = []
        # This will get updated as placement results are added to the
        # summary.
        self._num_atoms = num_atoms

        for id_, (building_block, result) in enumerate(
            zip(building_blocks, placement_results),
            num_previous_placements,
        ):
            self._with_placement_result(building_block, id_, result)
            self._num_atoms += building_block.get_num_atoms()

    def _with_placement_result(
        self,
        building_block,
        building_block_id,
        result,
    ):
        """
        Add the placement result to the summary.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block which was placed.

        building_block_id : :class:`int`
            A unique for the building block placement.

        result : :class:`._PlacementResult`
            The result of the placement.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._position_matrices.append(result.position_matrix)

        atom_batch = _AtomBatch(
            atoms=building_block.get_atoms(),
            num_atoms=self._num_atoms,
            building_block=building_block,
            building_block_id=building_block_id,
        )
        self._with_atom_batch(atom_batch)
        atom_map = atom_batch.get_atom_map()

        bond_batch = _BondBatch(
            bonds=building_block.get_bonds(),
            atom_map=atom_map,
            building_block=building_block,
            building_block_id=building_block_id,
        )
        self._with_bond_batch(bond_batch)

        self._with_functional_group_edges(
            building_block=building_block,
            functional_group_edges=result.functional_group_edges,
            atom_map=atom_map,
        )

    def _with_atom_batch(self, batch):
        """
        Add a batch of atoms to the summary.

        Parameters
        ----------
        batch : :class:`._AtomBatch`
            A batch of atoms.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._atoms.extend(batch.get_atoms())
        self._atom_infos.extend(batch.get_atom_infos())

    def _with_bond_batch(self, batch):
        """
        Add a batch of bonds to the summary.

        Parameters
        ----------
        batch : :class:`.BondBatch`
            A batch of bonds.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._bonds.extend(batch.get_bonds())
        self._bond_infos.extend(batch.get_bond_infos())

    def _with_functional_group_edges(
        self,
        building_block,
        functional_group_edges,
        atom_map,
    ):
        """
        Add the mapping from functional groups to edges.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block which owns the functional groups.

        functional_group_edges : :class:`dict`
            Maps the id of each functional group of `building_block`
            the the id of an edge.

        atom_map : :class:`dict`
            Maps the ids of atoms in `building_block` to the new atoms
            of the molecule being constructed.

        Returns
        -------
        None : :class:`NoneType`

        """

        functional_groups = building_block.get_functional_groups(
            fg_ids=functional_group_edges,
        )
        edge_ids = functional_group_edges.values()
        functional_group_edges = zip(functional_groups, edge_ids)
        for functional_group, edge_id in functional_group_edges:
            self._edge_functional_groups[edge_id].append(
                functional_group.with_atoms(atom_map)
            )

    def get_atoms(self):
        """
        Yield the atoms in the summary.

        Yields
        ------
        :class:`.Atom`
            An atom.

        """

        yield from self._atoms

    def get_atom_infos(self):
        """
        Yield infos about atoms in the summary.

        Yields
        ------
        :class:`.AtomInfo`
            Info about an atom.

        """

        yield from self._atom_infos

    def get_bonds(self):
        """
        Yield the bonds in the summary.

        Yields
        ------
        :class:`.Bond`
            A bond.

        """

        yield from self._bonds

    def get_bond_infos(self):
        """
        Yield infos about the bonds in the summary.

        Yields
        ------
        :class:`.BondInfo`
            Info about a bond.

        """

        yield from self._bond_infos

    def get_position_matrix(self):
        """
        Get a position matrix for the atoms in the summary.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix.

        """

        return np.vstack(self._position_matrices)

    def get_edge_functional_groups(self):
        """
        Yield the edge ids and functional groups associated with them.

        Yields
        ------
        :class:`tuple`
            Holds the edge id as the first element, and a
            :class:`tuple` of :class:`.FunctionalGroup` instances
            as the second element.

        """

        yield from self._edge_functional_groups.items()



from typing import NamedTuple


class _PlacementResult(NamedTuple):
    """
    The result of a building block placement.

    Attributes
    ----------
    position_matrix : :class:`numpy.ndarray`
        The position matrix of the building block after it has been
        placed on the vertex.

    functional_group_edges : :class:`dict`
        Maps the id of a functional group to the id of the edge it is
        assigned to.

    """

    position_matrix: object
    functional_group_edges: object


class _Placement:
    """
    Represents placement of a building block on a vertex.

    It represents a computation which carries out the placement
    of the building block on the vertex and the mapping of its
    functional group to edges.

    """

    def __init__(self, vertex, edges, building_block):
        """
        Initialize a :class:`._Placement`.

        Parameters
        ----------
        vertex : :class:`.Vertex`
            The vertex which does the placement.

        edges : :class:`tuple` of :class:`.Edge`
            The edges connected to `vertex`.

        building_block : :class:`.BuildingBlock`
            The building block to be placed on `vertex`.

        """

        self._vertex = vertex
        self._edges = edges
        self._building_block = building_block

    def get_result(self):
        """
        Get the result of the placement.

        Returns
        -------
        :class:`_PlacementResult`
            The result of the placement.

        """

        position_matrix = self._vertex.place_building_block(
            building_block=self._building_block,
            edges=self._edges,
        )
        position_matrix.setflags(write=False)
        building_block = self._building_block.with_position_matrix(
            position_matrix=position_matrix,
        )
        functional_group_edges = (
            self._vertex.map_functional_groups_to_edges(
                building_block=building_block,
                edges=self._edges,
            )
        )
        return _PlacementResult(
            position_matrix=position_matrix,
            functional_group_edges=functional_group_edges,
        )



"""
Property Vector
===============

"""

from .fitness_calculator import FitnessCalculator


class PropertyVector(FitnessCalculator):
    """
    Uses multiple molecular properties as a fitness value.

    Examples
    --------
    *Calculating Fitness Values*

    .. code-block:: python

        import stk

        # First, create the functions which calculate the properties
        # of molecules.
        def get_num_atoms(molecule):
            return molecule.get_num_atoms()

        def get_num_bonds(molecule):
            return molecule.get_num_bonds()

        def get_diameter(molecule):
            return molecule.get_maximum_diameter()

        # Next, create the fitness calculator.
        fitness_calculator = stk.PropertyVector(
            property_functions=(
                get_num_atoms,
                get_num_bonds,
                get_diameter,
            ),
        )

        # Calculate the fitness value of a molecule.
        # "value" is a tuple, holding the number of atoms, number of
        # bonds and the diameter of the molecule.
        value = fitness_calculator.get_fitness_value(
            molecule=stk.BuildingBlock('BrCCBr'),
        )

    *Storing Fitness Values in a Database*

    Sometimes you want to store fitness values in a database, you
    can do this by providing the `output_database` parameter.

    .. code-block:: python

        import stk
        # pymongo does not come with stk, you have to install it
        # separately with "pip install pymongo"
        import pymongo

        # Create a database which stores the fitness value of each
        # molecule.
        fitness_db = stk.ValueMongoDb(
            # This connects to a local database - so make sure you have
            # local MongoDB server running. You can also connect to
            # a remote MongoDB with MongoClient(), read to pymongo
            # docs to see how to do that.
            mongo_client=pymongo.MongoClient(),
            collection='fitness_values',
        )

        # Define the functions which calculate molecular properties.
        def get_num_atoms(molecule):
            return molecule.get_num_atoms()

        def get_num_bonds(molecule):
            return molecule.get_num_bonds()

        def get_diameter(molecule):
            return molecule.get_maximum_diameter()

        # Create the fitness calculator.
        fitness_calculator = stk.PropertyVector(
            property_functions=(
                get_num_atoms,
                get_num_bonds,
                get_diameter,
            ),
            output_database=fitness_db,
        )

        # Calculate fitness values.
        value1 = fitness_calculator.get_fitness_value(
            molecule=stk.BuildingBlock('BrCCBr'),
        )

        # You can retrieve the fitness values from the database.
        value = fitness_db.get(stk.BuildingBlock('BrCCBr'))

    *Caching Fitness Values*

    Usually, if you calculate the fitness value of a molecule, you
    do not want to re-calculate it, because this may be expensive,
    and the fitness value is going to be the
    same anyway. By using the `input_database` parameter, together
    with the `output_database` parameter, you can make sure you store
    and retrieve calculated fitness values instead of repeating the
    same calculation multiple times.

    The `input_database` is checked before a calculation happens, to
    see if the value already exists, while the `output_database` has
    the calculated fitness value deposited into it.

    .. code-block:: python

        import stk

        # pymongo does not come with stk, you have to install it
        # separately with "pip install pymongo"
        import pymongo

        # You can use the same database for both the input_database
        # and output_database parameters.
        fitness_db = stk.ValueMongoDb(
            # This connects to a local database - so make sure you have
            # local MongoDB server running. You can also connect to
            # a remote MongoDB with MongoClient(), read to pymongo
            # docs to see how to do that.
            mongo_client=pymongo.MongoClient(),
            collection='fitness_values',
        )

        # Define the functions which calculate molecular properties.
        def get_num_atoms(molecule):
            return molecule.get_num_atoms()

        def get_num_bonds(molecule):
            return molecule.get_num_bonds()

        def get_diameter(molecule):
            return molecule.get_maximum_diameter()

        # Create the fitness calculator.
        fitness_calculator = stk.PropertyVector(
            property_functions=(
                get_num_atoms,
                get_num_bonds,
                get_diameter,
            ),
            input_database=fitness_db,
            output_database=fitness_db,
        )

        # Assuming that a fitness value for this molecule was not
        # deposited into the database in a previous session, this
        # will calculate the fitness value.
        value1 = fitness_calculator.get_fitness_value(
            molecule=stk.BuildingBlock('BrCCBr'),
        )
        # This will not re-calculate the fitness value, instead,
        # value1 will be retrieved from the database.
        value2 = fitness_calculator.get_fitness_value(
            molecule=stk.BuildingBlock('BrCCBr'),
        )

    """

    def __init__(
        self,
        property_functions,
        input_database=None,
        output_database=None,
    ):

        """
        Initialize a :class:`.PropertyVector` instance.

        Parameters
        ----------
        property_fns : :class:`tuple` of :class:`callable`
            A group of :class:`function`, each of which is used to
            calculate a single property of the molecule. Each function
            must take one parameter, `mol`, which accepts
            a :class:`.Molecule` object. This is the molecule used to
            calculate the property.

        input_database : :class:`.ValueDatabase`, optional
            A database to check before calling `fitness_function`. If a
            fitness value exists for a molecule in the database, the
            stored value is returned, instead of calling
            `fitness_function`.

        output_database : :class:`.ValueDatabase`, optional
            A database into which the calculate fitness value is
            placed.

        """

        self._property_functions = property_functions
        self._input_database = input_database
        self._output_database = output_database

    def get_fitness_value(self, molecule):
        return tuple(
            property_function(molecule)
            for property_function in self._property_functions
        )



"""
Functional Group Factory
========================

.. toctree::
    :maxdepth: 2

    Alcohol Factory <\
stk.molecular.functional_groups.factories.alcohol_factory\
>
    Aldehyde Factory <\
stk.molecular.functional_groups.factories.aldehyde_factory\
>
    Amide Factory <\
stk.molecular.functional_groups.factories.amide_factory\
>
    Boronic Acid Factory <\
stk.molecular.functional_groups.factories.boronic_acid_factory\
>
    Bromo Factory <\
stk.molecular.functional_groups.factories.bromo_factory\
>
    Carboxylic Acid Factory <\
stk.molecular.functional_groups.factories.carboxylic_acid_factory\
>
    Dibromo Factory <\
stk.molecular.functional_groups.factories.dibromo_factory\
>
    Difluoro Factory <\
stk.molecular.functional_groups.factories.difluoro_factory\
>
    Diol Factory <\
stk.molecular.functional_groups.factories.diol_factory\
>
    Fluoro Factory <\
stk.molecular.functional_groups.factories.fluoro_factory\
>
    Iodo Factory <\
stk.molecular.functional_groups.factories.iodo_factory\
>
    Primary Amino Factory <\
stk.molecular.functional_groups.factories.primary_amino_factory\
>
    Ring Amine Factory <\
stk.molecular.functional_groups.factories.ring_amine_factory\
>
    Secondary Amino Factory <\
stk.molecular.functional_groups.factories.secondary_amino_factory\
>
    SMARTS Functional Group Factory <\
stk.molecular.functional_groups.factories.smarts_functional_group_factory\
>
    Terminal Alkene Factory <\
stk.molecular.functional_groups.factories.terminal_alkene_factory\
>
    Terminal Alkyne Factory <\
stk.molecular.functional_groups.factories.terminal_alkyne_factory\
>
    Thioacid Factory <\
stk.molecular.functional_groups.factories.thioacid_factory\
>
    Thiol Factory <\
stk.molecular.functional_groups.factories.thiol_factory\
>

"""


class FunctionalGroupFactory:
    """
    An abstract base class for functional group factories.

    The purpose of a functional group factory is to create
    :class:`.FunctionalGroup` instances. It allows the user to avoid
    creating :class:`.FunctionalGroup` instances manually. The user
    will use :meth:`.get_functional_groups` to achieve this. Subclasses
    of this class are made to customize the automatic creation
    process.

    Examples
    --------
    *Using a Single Functional Group From a Factory*

    You have a building block with two functional groups, but you want
    to use just one.

    .. code-block:: python

        import stk

        amino_factory = stk.PrimaryAminoFactory()
        building_block = stk.BuildingBlock('NCCN')
        amino_group, = amino_factory.get_functional_groups(
            molecule=building_block,
        )
        building_block = building_block.with_functional_groups(
            functional_groups=(amino_group, ),
        )

    *Using a Subset of Functional Groups From a Factory*

    You have multiple functional groups, but you want the building
    block to use a specific subset.

    .. code-block:: python

        import stk

        bromo_factory = stk.BromoFactory()
        building_block = stk.BuildingBlock('BrCC(Br)CC(Br)CC(Br)CCBr')
        bromo_groups = tuple(
            bromo_factory.get_functional_groups(molecule)
        )
        building_block = building_block.with_functional_groups(
            functional_groups=(
                bromo_groups[0],
                bromo_groups[3],
                bromo_groups[4],
            ),
        )

    More usage examples can be found in the docstrings of the
    various subclasses.

    *Subclass Implementation*

    The source of the subclasses, listed in
    :mod:`.functional_group_factory`, can serve as good examples.

    """

    def get_functional_groups(self, molecule):
        """
        Yield functional groups in `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule, whose functional groups are to be found.

        Yields
        ------
        :class:`.FunctionalGroup`
            A functional group in `molecule`.

        Examples
        --------
        See :class:`.FunctionalGroupFactory`.

        """

        raise NotImplementedError()



"""
Boronic Acid
============

"""

from .generic_functional_group import GenericFunctionalGroup


class BoronicAcid(GenericFunctionalGroup):
    """
    Represents a boronic acid functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[atom][boron]([oxygen1][hydrogen1])[oxygen2][hydrogen2]``.

    """

    def __init__(
        self,
        boron,
        oxygen1,
        hydrogen1,
        oxygen2,
        hydrogen2,
        atom,
        bonders,
        deleters,
    ):
        """
        Initialize a :class:`.BoronicAcid` instance.

        Parameters
        ----------
        boron : :class:`.B`
            The ``[boron]`` atom.

        oxygen1 : :class:`.O`
            The ``[oxygen1]`` atom.

        hydrogen1 : :class:`.H`
            The ``[hydrogen]`` atom.

        oxygen2 : :class:`.O`
            The ``[oyxgen2]`` atom.

        hydrogen2 : :class:`.H`
            The ``[hydrogen2]`` atom.

        atom : :class:`.Atom`
            The ``[atom]`` atom.

        bonders : :class:`tuple` of :class:`.Atom`
            the bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        """

        self._boron = boron
        self._oxygen1 = oxygen1
        self._hydrogen1 = hydrogen1
        self._oxygen2 = oxygen2
        self._hydrogen2 = hydrogen2
        self._atom = atom
        atoms = (boron, oxygen1, hydrogen1, oxygen2, hydrogen2, atom)
        super().__init__(atoms, bonders, deleters)

    def get_boron(self):
        """
        Get the ``[boron]`` atom.

        Returns
        -------
        :class:`.B`
            The ``[boron]`` atom.

        """

        return self._boron

    def get_oxygen1(self):
        """
        Get the ``[oxygen1]`` atom.

        Returns
        -------
        :class:`.O`
            The ``[oxygen1]`` atom.

        """

        return self._oxygen1

    def get_hydrogen1(self):
        """
        Get the ``[hydrogen1]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen1]`` atom.

        """

        return self._hydrogen1

    def get_oxygen2(self):
        """
        Get the ``[oxygen2]`` atom.

        Returns
        -------
        :class:`.O`
            The ``[oxygen2]`` atom.

        """

        return self._oxygen2

    def get_hydrogen2(self):
        """
        Get the ``[hydrogen2]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen2]`` atom.

        """

        return self._hydrogen2

    def get_atom(self):
        """
        Get the ``[atom]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom]`` atom.

        """

        return self._atom

    def clone(self):
        clone = super().clone()
        clone._boron = self._boron
        clone._oxygen1 = self._oxygen1
        clone._hydrogen1 = self._hydrogen1
        clone._oxygen2 = self._oxygen2
        clone._hydrogen2 = self._hydrogen2
        clone._atom = self._atom
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._boron = atom_map.get(
            self._boron.get_id(),
            self._boron,
        )
        clone._oxygen1 = atom_map.get(
            self._oxygen1.get_id(),
            self._oxygen1,
        )
        clone._hydrogen1 = atom_map.get(
            self._hydrogen1.get_id(),
            self._hydrogen1,
        )
        clone._oxygen2 = atom_map.get(
            self._oxygen2.get_id(),
            self._oxygen2,
        )
        clone._hydrogen2 = atom_map.get(
            self._hydrogen2.get_id(),
            self._hydrogen2,
        )
        clone._atom = atom_map.get(
            self._atom.get_id(),
            self._atom,
        )
        return clone

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'{self._boron}, {self._oxygen1}, {self._hydrogen1}, '
            f'{self._oxygen2}, {self._hydrogen2}, {self._atom}, '
            f'bonders={self._bonders}, deleters={self._deleters})'
        )



"""
Secondary Amino Factory
=======================

"""

from .functional_group_factory import FunctionalGroupFactory
from .utilities import _get_atom_ids
from ..functional_groups import SecondaryAmino


class SecondaryAminoFactory(FunctionalGroupFactory):
    """
    Creates :class:`.SecondaryAmino` instances.

    Creates functional groups from substructures, which match the
    ``[H][N]([#6])[#6]`` functional group string.

    Examples
    --------
    You want to create a building block which has
    :class:`.SecondaryAmino` functional groups. You want the nitrogen
    atom in those functional groups to be the *bonder* atom, and the
    hydrogen atom to be the *deleter* atom.

    .. code-block:: python

        import stk

        building_block = stk.BuildingBlock(
            smiles='HN(CCCC)CCCC',
            functional_groups=(stk.SecondaryAminoFactory(), ),
        )

    You want to create a building block which has
    :class:`.SecondaryAmino` functional groups. You want the nitrogen
    atom to be the *bonder* atom and one of the carbon atoms to be the
    *deleter* atom.

    .. code-block:: python

        import stk

        secondary_amino_factory = stk.SecondaryAminoFactory(
            # The index of the nitrogen atom in the functional
            # group string (see docstring) is 1.
            bonders=(1, ),
            # The index of one of the carbon atoms in the functional
            # group string (see docstring) is 2.
            deleters=(2, ),
        )
        building_block = stk.BuildingBlock(
            smiles='HN(CCCC)CCCC',
            functional_groups=(secondary_amino_factory, ),
        )


    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(1, ), deleters=(0, )):
        """
        Initialize a :class:`.SecondaryAminoFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are bonder atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are deleter atoms.

        """

        self._bonders = bonders
        self._deleters = deleters

    def get_functional_groups(self, molecule):
        for atom_ids in _get_atom_ids('[H][N]([#6])[#6]', molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield SecondaryAmino(
                nitrogen=atoms[1],
                hydrogen=atoms[0],
                atom1=atoms[2],
                atom2=atoms[3],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
            )



"""
Cage
====

.. toctree::
    :maxdepth: 2

    One Plus One <\
stk.molecular.topology_graphs.cage.three_plus_three.one_plus_one\
>
    Two Plus Two <\
stk.molecular.topology_graphs.cage.three_plus_three.two_plus_two\
>
    Two Plus Three <\
stk.molecular.topology_graphs.cage.two_plus_three.two_plus_three\
>
    Two Plus Four <\
stk.molecular.topology_graphs.cage.two_plus_four.two_plus_four\
>
    Three Plus Six <\
stk.molecular.topology_graphs.cage.two_plus_four.three_plus_six\
>
    Four Plus Four <\
stk.molecular.topology_graphs.cage.three_plus_three.four_plus_four\
>
    Four Plus Six <\
stk.molecular.topology_graphs.cage.two_plus_three.four_plus_six\
>
    Four Plus Six 2 <\
stk.molecular.topology_graphs.cage.two_plus_three.four_plus_six_2\
>
    Four Plus Eight <\
stk.molecular.topology_graphs.cage.two_plus_four.four_plus_eight\
>
    Five Plus Ten <\
stk.molecular.topology_graphs.cage.two_plus_four.five_plus_ten\
>
    Six Plus Eight <\
stk.molecular.topology_graphs.cage.three_plus_four.six_plus_eight\
>
    Six Plus Nine <\
stk.molecular.topology_graphs.cage.two_plus_three.six_plus_nine\
>
    Six Plus Twelve <\
stk.molecular.topology_graphs.cage.two_plus_four.six_plus_twelve\
>
    Eight Plus Twelve <\
stk.molecular.topology_graphs.cage.two_plus_three.eight_plus_twelve\
>
    Eight Plus Sixteen <\
stk.molecular.topology_graphs.cage.two_plus_four.eight_plus_sixteen\
>
    Ten Plus Twenty <\
stk.molecular.topology_graphs.cage.two_plus_four.ten_plus_twenty\
>
    Twelve Plus Thirty <\
stk.molecular.topology_graphs.cage.two_plus_five.twelve_plus_thirty\
>
    Twenty Plus Thirty <\
stk.molecular.topology_graphs.cage.two_plus_three.twenty_plus_thirty\
>

"""

from collections import Counter, defaultdict
from functools import partial

from .cage_construction_state import _CageConstructionState
from ..topology_graph import TopologyGraph
from ...reactions import GenericReactionFactory


class UnoccupiedVertexError(Exception):
    """
    When a cage vertex is not occupied by a building block.

    """

    pass


class OverlyOccupiedVertexError(Exception):
    """
    When a cage vertex is occupied by more than one building block.

    """

    pass


class Cage(TopologyGraph):
    """
    Represents a cage topology graph.

    Notes
    -----
    Cage topologies are added by creating a subclass, which defines the
    :attr:`_vertex_prototypes` and :attr:`_edge_prototypes` class
    attributes.

    Examples
    --------
    *Subclass Implementation*

    The source code of the subclasses, listed in :mod:`~.cage.cage`,
    can serve as good examples.

    *Basic Construction*

    :class:`.Cage` instances can be made by providing the building
    block molecules only (using :class:`.FourPlusSix` as an example)

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='O=CC(C=O)C=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        cage1 = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusSix((bb1, bb2)),
        )

    *Structural Isomer Construction*

    Different structural isomers of cages can be made by using the
    `vertex_alignments` optional parameter

    .. code-block:: python

        cage2 = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusSix(
                building_blocks=(bb1, bb2),
                vertex_alignments={0: 1, 1: 1, 2: 2},
            ),
        )

    The parameter maps the id of a vertex to a number
    between 0 (inclusive) and the number of edges the vertex is
    connected to (exclusive). So a vertex connected to three edges
    can be mapped to ``0``, ``1`` or ``2``.

    By changing which edge each vertex is aligned with, a different
    structural isomer of the cage can be formed.

    *Multi-Building Block Cage Construction*

    You can also build cages with multiple building blocks, but,
    if you have multiple building blocks with the same number
    of functional groups, you have to assign each building block to the
    vertex you want to place it on

    .. code-block:: python

        bb1 = stk.BuildingBlock(
            smiles='O=CC(C=O)C=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='O=CC(Cl)(C=O)C=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        bb3 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb4 = stk.BuildingBlock(
            smiles='NCC(Cl)N',
            functional_groups=[stk.PrimaryAminoFactory()],
        )
        bb5 = stk.BuildingBlock('NCCCCN', [stk.PrimaryAminoFactory()])

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusSix(
                # building_blocks is now a dict, which maps building
                # blocks to the id of the vertices it should be placed
                # on. You can use ranges to specify the ids.
                building_blocks={
                    bb1: range(2),
                    bb2: (2, 3),
                    bb3: 4,
                    bb4: 5,
                    bb5: range(6, 10),
                },
            ),
        )

    You can combine this with the `vertex_alignments` parameter

    .. code-block:: python

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusSix(
                building_blocks={
                    bb1: range(2),
                    bb2: (2, 3),
                    bb3: 4,
                    bb4: 5,
                    bb5: range(6, 10),
                },
            ),
            vertex_alignments={0: 1, 1: 1, 2: 2},
        )

    """

    def __init_subclass__(cls, **kwargs):
        cls._vertex_degrees = Counter(
            vertex_id
            for edge in cls._edge_prototypes
            for vertex_id in edge.get_vertex_ids()
        )
        cls._vertices_of_degree = defaultdict(set)
        for vertex_id, degree in cls._vertex_degrees.items():
            cls._vertices_of_degree[degree].add(vertex_id)

    def __init__(
        self,
        building_blocks,
        vertex_alignments=None,
        reaction_factory=GenericReactionFactory(),
        num_processes=1,
    ):
        """
        Initialize a :class:`.Cage`.

        Parameters
        ----------
        building_blocks : :class:`iterable` or :class:`dict`
            Can be a :class:`iterable` of :class:`.BuildingBlock`
            instances, which should be placed on the topology graph.

            Can also be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on. A :class:`dict` is
            required when there are multiple building blocks with the
            same number of functional groups, because in this case
            the desired placement is ambiguous.

        vertex_alignments : :class:`dict`, optional
            A mapping from the id of a :class:`.Vertex`
            to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is referred to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        reaction_factory : :class:`.ReactionFactory`, optional
            The reaction factory to use for creating bonds between
            building blocks.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        Raises
        ------
        :class:`AssertionError`
            If the any building block does not have a
            valid number of functional groups.

        :class:`ValueError`
            If the there are multiple building blocks with the
            same number of functional_groups in `building_blocks`,
            and they are not explicitly assigned to vertices. The
            desired placement of building blocks is ambiguous in
            this case.

        :class:`~.cage.UnoccupiedVertexError`
            If a vertex of the cage topology graph does not have a
            building block placed on it.

        :class:`~.cage.OverlyOccupiedVertexError`
            If a vertex of the cage topology graph has more than one
            building block placed on it.

        """

        # Use tuple here because it prints nicely.
        allowed_degrees = tuple(self._vertices_of_degree.keys())
        if isinstance(building_blocks, dict):
            for building_block in building_blocks:
                assert (
                    building_block.get_num_functional_groups()
                    in self._vertices_of_degree.keys()
                ), (
                    'The number of functional groups in '
                    f'{building_block} needs to be one of '
                    f'{allowed_degrees}, but is '
                    'currently '
                    f'{building_block.get_num_functional_groups()}.'
                )
            building_blocks = {
                building_block: self._get_vertices(ids)
                for building_block, ids in building_blocks.items()
            }

        else:
            building_blocks = self._get_building_block_vertices(
                building_blocks=building_blocks,
            )

        self._vertex_alignments = vertex_alignments = (
            dict(vertex_alignments)
            if vertex_alignments is not None
            else {}
        )

        def with_aligner(vertex):
            return vertex.with_aligner_edge(
                aligner_edge=vertex_alignments.get(vertex.get_id(), 0),
            )

        building_block_vertices = {
            building_block: tuple(map(with_aligner, vertices))
            for building_block, vertices in building_blocks.items()
        }
        self._check_building_block_vertices(building_block_vertices)
        super().__init__(
            building_block_vertices=building_block_vertices,
            edges=self._edge_prototypes,
            reaction_factory=reaction_factory,
            construction_stages=tuple(
                partial(self._has_degree, degree)
                for degree
                in sorted(self._vertices_of_degree, reverse=True)
            ),
            num_processes=num_processes,
            edge_groups=None,
        )

    @classmethod
    def _check_building_block_vertices(cls, building_block_vertices):
        unassigned_ids = set(
            vertex.get_id() for vertex in cls._vertex_prototypes
        )
        assigned_ids = set()
        vertices = (
            vertex
            for vertices_ in building_block_vertices.values()
            for vertex in vertices_
        )
        for vertex in vertices:
            if vertex.get_id() in assigned_ids:
                raise OverlyOccupiedVertexError(
                    f'Vertex {vertex.get_id()} has multiple building '
                    'blocks placed on it.'
                )
            assigned_ids.add(vertex.get_id())
            unassigned_ids.remove(vertex.get_id())

        if unassigned_ids:
            raise UnoccupiedVertexError(
                'The following vertices are unoccupied '
                f'{unassigned_ids}.'
            )

    def clone(self):
        clone = super().clone()
        clone._vertex_alignments = dict(self._vertex_alignments)
        return clone

    def _get_vertices(self, vertex_ids):
        """
        Yield vertex prototypes.

        Parameters
        ----------
        vertex_ids : :class:`iterable` of :class:`int`
            The ids of the vertices to yield.

        Yields
        ------
        :class:`.Vertex`
            A vertex prototype of the topology graph.

        """

        if isinstance(vertex_ids, int):
            vertex_ids = (vertex_ids, )

        for vertex_id in vertex_ids:
            yield self._vertex_prototypes[vertex_id]

    def _has_degree(self, degree, vertex):
        """
        Check if `vertex` has a degree of `degree`.

        Parameters
        ----------
        degree : :class:`int`
            The degree in question.

        vertex : :class:`.Vertex`
            The vertex in question.

        Returns
        -------
        :class:`bool`
            ``True`` if `vertex` has a degree of `degree`.

        """

        return vertex.get_id() in self._vertices_of_degree[degree]

    def _get_building_block_vertices(self, building_blocks):
        """
        Map building blocks to the vertices of the graph.

        Parameters
        ----------
        building_blocks : :class:`iterable` of :class:`.BuildingBlock`
            The building blocks which need to be mapped to vertices.

        Returns
        -------
        :class:`dict`
            Maps each building block in `building_blocks` to a
            :class:`list` of :class:`.Vertex` instances it should be
            placed on.

        Raises
        ------
        :class:`AssertionError`
            If the any building block does not have a
            valid number of functional groups.

        :class:`ValueError`
            If there are multiple building blocks with the same number
            of functional groups.

        """

        # Use tuple here because it prints nicely.
        allowed_degrees = tuple(self._vertices_of_degree.keys())

        building_blocks_by_degree = {}
        for building_block in building_blocks:
            num_fgs = building_block.get_num_functional_groups()
            assert (
                num_fgs in self._vertices_of_degree.keys()
            ), (
                'The number of functional groups in '
                f'{building_block} needs to be one of '
                f'{allowed_degrees}, but is '
                'currently '
                f'{building_block.get_num_functional_groups()}.'
            )

            if num_fgs in building_blocks_by_degree:
                raise ValueError(
                    'If there are multiple building blocks with the '
                    'same number of functional groups, '
                    'building_block_vertices must be set explicitly.'
                )
            building_blocks_by_degree[num_fgs] = building_block

        building_block_vertices = {}
        for vertex in self._vertex_prototypes:
            vertex_degree = self._vertex_degrees[vertex.get_id()]
            building_block = building_blocks_by_degree[vertex_degree]
            building_block_vertices[building_block] = (
                building_block_vertices.get(building_block, [])
            )
            building_block_vertices[building_block].append(vertex)
        return building_block_vertices

    def _get_scale(self, building_block_vertices):
        return max(
            bb.get_maximum_diameter()
            for bb in building_block_vertices
        )

    def _get_construction_state(self):
        return _CageConstructionState(
            building_block_vertices=self._building_block_vertices,
            edges=self._edges,
            num_placement_stages=self._implementation.get_num_stages(),
            vertex_degrees=self._vertex_degrees,
        )

    def __repr__(self):
        vertex_alignments = (
            ', vertex_alignments={self._vertex_alignments}'
            if self._vertex_alignments
            else ''
        )
        return f'cage.{self.__class__.__name__}({vertex_alignments})'



"""
Six Plus Nine
=============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class SixPlusNine(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    # source: http://eusebeia.dyndns.org/4d/prism3
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [-1, -1/np.sqrt(3), -1]),
        _NonLinearCageVertex(1, [-1, -1/np.sqrt(3), 1]),
        _NonLinearCageVertex(2, [1, -1/np.sqrt(3), -1]),
        _NonLinearCageVertex(3, [1, -1/np.sqrt(3), 1]),
        _NonLinearCageVertex(4, [0, 2/np.sqrt(3), -1]),
        _NonLinearCageVertex(5, [0, 2/np.sqrt(3), 1])
    )
    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCageVertex.init_at_center(
            id=6,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
        ),
        _LinearCageVertex.init_at_center(
            id=7,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
        ),
        _LinearCageVertex.init_at_center(
            id=8,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[3]),
        ),
        _LinearCageVertex.init_at_center(
            id=9,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[3]),
        ),
        _LinearCageVertex.init_at_center(
            id=10,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[4]),
        ),
        _LinearCageVertex.init_at_center(
            id=11,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[4]),
        ),
        _LinearCageVertex.init_at_center(
            id=12,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[5]),
        ),
        _LinearCageVertex.init_at_center(
            id=13,
            vertices=(_vertex_prototypes[3], _vertex_prototypes[5]),
        ),
        _LinearCageVertex.init_at_center(
            id=14,
            vertices=(_vertex_prototypes[4], _vertex_prototypes[5]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[6], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[6], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[7], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[7], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[8], _vertex_prototypes[2]),
        Edge(5, _vertex_prototypes[8], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[9], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[9], _vertex_prototypes[3]),
        Edge(8, _vertex_prototypes[10], _vertex_prototypes[0]),
        Edge(9, _vertex_prototypes[10], _vertex_prototypes[4]),
        Edge(10, _vertex_prototypes[11], _vertex_prototypes[2]),
        Edge(11, _vertex_prototypes[11], _vertex_prototypes[4]),
        Edge(12, _vertex_prototypes[12], _vertex_prototypes[1]),
        Edge(13, _vertex_prototypes[12], _vertex_prototypes[5]),
        Edge(14, _vertex_prototypes[13], _vertex_prototypes[3]),
        Edge(15, _vertex_prototypes[13], _vertex_prototypes[5]),
        Edge(16, _vertex_prototypes[14], _vertex_prototypes[4]),
        Edge(17, _vertex_prototypes[14], _vertex_prototypes[5]),
    )

    _num_windows = 5
    _num_window_types = 1
